
#ifndef __tr__
#define __tr__

#include "db.h"
#include "stock.h"
#include "msd.h"
#include "steim.h"

#ifdef	__cplusplus
extern "C" {
#endif

typedef float Trsample ;

/* tr data codes */
#define trInternalInt   1
#define trInternalShort 2
#define trInternalFloat 3
#define trFLOAT  	4
#define trSHORT  	5
#define trINT	 	6
#define trSEED   	7
#define trGAIN_RANGE 	8
#define trIDA_2BYTE 	9
#define trIDA_4BYTE 	10
#define trASCII		11
#define trINTELFLOAT	12
#define trINTELSHORT	13
#define trINTELINT	14
#define trDISCARD	15
#define trGSE_CM6	16
#define trSAC		19
#define trAH		20
#define trAAH		21
#define trINTEL3BYTE    22
#define tr3BYTE		23
#define trREFTEK	24
#define trISAC		25
#define trCA		26
#define trFLATLINE	27
#define trSTEIM1   	28

#define trByCHAN 	1
#define trBySTA  	2
#define trByEVENT  	3
#define trByNETWORK	4

/* bits returned by i/tr2ext() */
#define TR_CLIPPED      2	/* indicates output data type doesn't have enough bits to represent internal data */
#define TR_TRUNCATED    1	/* indicates non-integer input */
#define TR_GAPS		8	/* indicates marked gap in data */

/* trsave(3) bits in flag */
#define trOVERWRITE 	 (1<<0)	/* overwrite existing waveform file */
#define trAPPEND 	 (1<<1)	/* append to any existing waveform file */
#define trAVOID_OVERLAP  (1<<2)	/* require that database row not conflict 
					with existing rows */
#define trUSE_LOCKS      (1<<3)   /* lock output files */
#define trFLUSH	         (1<<4)   /* flush output files */
#define trOLD_WFFILE	 (1<<5)	  /* use existing dir/dfile inside trwfnew */
#define trOLD_WFDISC	 (1<<6)	  /* overwrite existing wfdisc record inside trwfnew */
#define trNEW_WFFILE	 (1<<7)	  /* generate new wfname inside trwfnew if previous already exists */

/* structure to hold fill values, and representation limits for */
/* different datatypes */
typedef struct Wftype { 
    char 	datatype[4] ; 	/* datatype following wfdisc.datatype conventions */
    int	        code ; 		/* corresponding integer value for use in programs */
    double	fill ; 		/* default fill value to indicate missing data */
    double 	lower ; 	/* test for missing value is: */
    double 	upper ;    	/*   value <= lower || value >= upper */
    int	        bytesPerSamp ;  /* # bytes/sample if constant, otherwise zero */
    char *wfin ;		/* name of routine to read input data */
    char *wfwhdr ;		/* name of routine to write output header */
    char *wfout ;		/* name of routine to write output data */
    char *wftail ;		/* name of routine to write output tail (checksum for GSE_CM6) */
    int  appendable ; 		/* non-zero if wfout may be called repeatedly. */
    char *desc ;		/* short description of data type */
} Wftype ;

typedef struct Trsegment
{
    int             first,
                    last;
}               Trsegment;

typedef struct Trbuffer
{
    Trsample *data ; 
    double t0, t1 ; 
    int	nsamp; 
    Dbptr db ; 
} Trbuffer ;
	    
typedef struct Trseg {
    char sta[25], chan[25] ;
    double time, endtime, samprate, calib ; 
    int nsamp ;
    Trsample *data ;
    Dbptr tr ;
} Trseg ;

#define SAMP2TIME(TIME,SAMPRATE,ISAMP) ((TIME) + ((ISAMP))  /(SAMPRATE))
#define ENDTIME(TIME,SAMPRATE,NSAMP)   ((TIME) + ((NSAMP)-1)/(SAMPRATE))
#define TIME2SAMP(TIME,SAMPRATE,TIME1) floor(((TIME1)  -  (TIME))*(SAMPRATE) + 0.5)
#define NSAMP(TIME,SAMPRATE,ENDTIME)   floor(((ENDTIME) - (TIME))*(SAMPRATE) + 1.5)
#define SAMPRATE(TIME,NSAMP,ENDTIME)   (((NSAMP)-1)/((ENDTIME)-(TIME)))

#define TRCONTIGUOUS(T0,T1,R0,N0)       (ABS(((T1)-(T0))*(R0)-(N0))< trTOLERANCE)
#define TRSAMERATE(R0,R1)       (ABS(1.0-(R0)/(R1)) < trSAMPRATE_TOLERANCE)
#define TRSAMETICKS(T0,T1,R0)   (ABS( (((T1)-(T0))*(R0))-(floor(((T1)-(T0))*(R0) + 0.5)) ) < trTOLERANCE)
#define TRSAMETIME(T0,T1,R0) 	((ABS((T0)-(T1))*(R0)) < trTOLERANCE)

#define TRBADSEED_VALUE (268435456.)
#define TRGAP_VALUE     268435456
#define TRGAP_VALUE_S4  2147483647
#define NSAMP_LIMIT 	99999999

#define DT_CSS          .000005	/* uncertainty in time from css tables */
#define DT_SEED         .00005  /* uncertainty in time from miniseed */
#define DT_TRUNC        (DT_CSS+DT_SEED)
#define MSDSAMETIME(T0,T1,R0) 	(((ABS((T0)-(T1))-DT_TRUNC)*(R0)) < trTOLERANCE)

typedef struct Wfdisc {
    char net[32] ;			/* network name : network and loc are used only in msd2wf */
    char sta[32] ;			/* station name */
    char chan[32] ;			/* channel name */
    char loc[32] ; 			/* loc code */
    double from, until ; 	        /* time range for output should be time <= t < until */
    double time ; 			/* starting time for segment */
    double samprate ; 			/* samprate in wfdisc */
    double calib ;			/* calib in wfdisc */
    double calper ;			/* calper in wfdisc */
    int nsamp ; 			/* current nsamp in output */
    int checksum ; 			/* cumulative checksum for GSE_CM6 output */
    char segtype[4] ; 			/* response type in wfdisc */
    char instype[8] ; 			/* instrument type in wfdisc */
    char *wfname ;			/* trwfname(3) pattern for output data files */
    Wftype *wftype ;			/* output data type */
    Msd *msd ;				/* miniseed initialization */
    Hook *hook ; 			/* hook for private info */
    off_t foff ; 				/* original offset for data in output file */
    char path[FILENAME_MAX] ;		/* pathname for output */
    char  dir[68] ; 
    char  dfile[36] ;
    int	  fd ; 				/* if > 0, open file descriptor */
    char *wf ; 				/* output buffer */
    int   wf_nbytes ;			/* # of bytes in output buffer */
    int   wf_nsamp ; 		        /* # of sample in wf output buffer */
    int   wf_size ;			/* size of output buffer when buffering data */
    Dbptr db ;				/* output database row */
    Dbptr dbhdr ;			/* db pointer from which to get lat, lon, etc for header */
    int flags ; 			/* flags for specifying behavior of writing routines */
} Wfdisc ;


extern int trMAX_OPEN_FILES ; 
extern int trVerbose_splicing ; 
extern double trTOLERANCE ;
extern double trSAMPRATE_TOLERANCE ;
extern int  Multiply_calib_in_sac_output ;
extern char *trDefault_datatype ; 
extern char *trDefault_schema ; 
extern char *trReference_coordinates ;
extern char Foreignkeys_database[256] ;
extern int trMax_open_files ;
extern Xlat Trxlat[] ;
extern int NTrxlat ;
extern char *trDefault_autodrm_network ;
extern char *trDefault_seed_network ;
extern char *trDefault_loc_code ;
extern char *trDefault_aux_code ;
extern char *Default_waveform_pattern ;
extern Tbl *trSeed_net_sta ;
extern Tbl *trSeed_sta_chan_loc ;
extern Tbl *trAutodrm_net_sta ;
extern Tbl *trAutodrm_sta_chan_aux ;
extern int Miniseed_dataformat ;
extern int Miniseed_record_size ;
extern int Miniseed_has_b100 ;
extern int Miniseed_has_b1001 ;
extern int Miniseed_sparc_order ;
extern int Miniseed_ignore_errors ;
extern int Miniseed_segment_seconds ;
extern int Miniseed_timing_quality ; 

extern Dbptr trdup ( Dbptr tr );
extern Dbptr trnew ( char *tr_name, char *schema_name );
extern Hook * bwt_new ( double fl, int ol, double fu, int ou, double dt );
extern int bwt_init ( Hook *hook, float xin[] );
extern int bwt_filter ( Hook *hook, int npts, float xin[], int isave, float xout[] );
extern int db2tr1 ( Dbptr db, Dbptr dbtr, char *table, Arr *expressions, Hook **hookp );
extern int samprate2seed ( double samprate, int *factor, int *multiplier );
extern int seed2samprate ( int factor, int multiplier, double *samprate );
extern int trapply ( Dbptr tr, int (*filter) (), void *);
extern int trapply_calib ( Dbptr tr );
extern int trcontiguous ( Trseg *prv, Trseg *cur, double tolerance );
extern int trdatacode ( char *datatype );
extern char * trdatatype ( int datacode );
extern int trdemean ( Dbptr tr );
extern int trdemean_seg ( Dbptr tr );
extern int trderiv ( Dbptr tr );
extern int trdestroy ( Dbptr *trp );
extern int trdisp ( Dbptr tr, char *title );
extern int trfake ( Dbptr tr, char *params );
extern int trfake_argv ( Dbptr tr, int argc, char **argv );
extern int trfilter (Dbptr tr, char *filter_string);
extern Tbl * trfilter_parse (char *filter_string);
extern int trfilter_segs (int nsegs, int *nsamps, double *dt, float **data, char *filter_string);
extern int trfilter_pkt (int nsamps, double dt, float *data, char *filter_string, Hook **hook);
extern int trfree ( Dbptr tr );
extern int trgetASCIIwf ( char *s, int nbytes, int samp0, int nrequest, float *dest );
extern int trgetIDAwf ( char *data, int nbytes, int samp0, int nrequest, float *dest );
extern int trgetSEEDwf ( unsigned char *seed, int size, double samprate, int nsamp, double wfdisc_time, double time0, int samp0, int nrequest, float *dest );
extern int trsplice ( Dbptr tr, double tolerance, int (*fill)(Dbptr *, Trsample *, int *, int *, int *), int (*resample)(Dbptr *, Trsample *, Trsample *, int, Dbptr *, Trsample *, int *, double *, double *) );
extern int trintegral ( Dbptr tr );
extern int trload_css ( Dbptr db, char *time_str, char *endtime_str, Dbptr *trp, char *table, Arr **maparr );
extern int trload_cssgrp ( Dbptr db, char *time_str, char *endtime_str, Dbptr *trp, char *table, Arr **maparr );
extern int trload_arrivals ( Dbptr db, char *time_str, char *endtime_str, Dbptr *trp, char *table );
extern int trloadwf ( Dbptr db, double t0, double t1, Dbptr tr );
extern Dbptr trloadchan ( Dbptr db, double t0, double t1, char *sta, char *chan );
extern int troverlap ( double req_start, double req_end, double time, double endtime, double samprate, int *pt0p, double *t0p, double *t1p, int *npts );
extern double traligned ( double atime, double stdtime, double stdsamprate);
extern int trputwf ( Dbptr db, float *data );
extern int trputwf_ ( Dbptr *f_db, float *data );
extern int trsamerate ( Trseg *prv, Trseg *cur, double tolerance );
extern int trsameticks ( Trseg *prv, Trseg *cur, double tolerance );
extern int trsample ( Dbptr tr, int i0, int n, int dn );
extern int trsave_wf ( Dbptr tr, Dbptr db, char *datatype, char *subdir, int flags );
extern int trsplit (Dbptr tr, int (*fill) (Dbptr *, float *, int *, int *, int * ), int (*findbad)(Trsample *, int, char *, Tbl *) );
extern int trwfname ( Dbptr db, char *pattern, char **pathp );
extern int trcontinuous (double time, int nsamp, double tnew,  double *samprate, double *min_rate, double *max_rate ) ;

extern int wfinG2 ( char *gp, int sz, int samp0, int n, float *f );
extern void trgetwf_ ( Dbptr *f_db, Arr **f_maparr, float *data, int *f_nmax, double *f_reqt0, double *f_reqt1, double *t0p, double *t1p, int *nptsp, int (*fill)(), void **);
extern int trgetwf ( Dbptr db, Arr **maparr, Trsample **data, int *datasz, double reqt0, double reqt1, double *t0p, double *t1p, int *nptsp, int (*fill)(), void *);
extern void trgetwf_error ( Dbptr db, int code );

#define trgetwf2 trgetwf
extern int trtruncate ( Dbptr tr, int leave );

extern int wf2disc ( void *data, int nbytes, int nsamp, Wfdisc *wfdisc ) ;
extern int wfnew ( Dbptr db, Wfdisc *wfdisc );
extern int wfchanid ( Dbptr db );
extern int arrchanid ( Dbptr db );
extern int wfwrite ( void *data, int nbytes, int nsamp, Wfdisc *wfdisc, int flush );

extern void trfill_gap ( Trsample *data, int nsamp );
extern void trfill_buf ( Wftype *type, int nsamp, void **out, int *sZout, int *outbytes );
extern void trfindgaps ( Trsample *data, int nsamp, Tbl *badlist );
extern void gaps2tr ( Trsample *data, int nsamp, char *datatype );
extern void tr2gaps ( Trsample *data, int nsamp, char *datatype );
extern int wfdisc2tr ( void *in, int nsamp, char *datatype, Trsample **out, int *sZout );
extern int tr2ext ( Trsample *in, int nsamp, Wftype *type, void **out, int *sZout, int *outbytes );
extern int itr2ext ( int *in, int nsamp, Wftype *type, void **out, int *sZout, int *outbytes );

extern int trsignal_argv ( Dbptr tr, int argc, char **argv );
extern int trsignal ( Dbptr tr, char *params );
extern void imksignal ( char *type, int *data, double time, int nsamp, double samprate, double t0, double freq, double slew, double amp );
extern void mksignal ( char *type, float *data, double time, int nsamp, double samprate, double t0, double freq, double slew, double amp );

extern int  rotate_to_standard(Dbptr tr,char **newchan);
extern int trrotate(Dbptr tr, double phi, double theta, char **newchan);

extern int trclip ( double time, double samprate, int nsamp, double from, double until, int *pt0p, int *nptsp, double *t0p, double *t1p ) ;

extern int trlookup_segtype ( char *segtype, char **segunits, char **segdesc );
extern Wftype * trwftype ( char *datatype );
extern Wftype * atrwftype ( char *datatype );
extern Wftype * trwflist ( int i );
extern void init_wfdisc ( Wfdisc *wfdisc, char *wfname, char *datatype, int bufsize, Dbptr dbout, Dbptr dbhdr, int flags);
extern Wfdisc * new_wfdisc ( char *wfname, char *wftype, int bufsize, Dbptr dbout, Dbptr dbhdr, int flags );
extern void free_wfdisc ( Wfdisc *wfdisc );

extern int trlockwf(int fd, char *path) ;
extern int trwfnew ( Wfdisc *wfdisc );
extern int trwfclose ( Wfdisc *wfdisc );

extern int wfin3BYTE ( char *buffer, int nbytes, int first, int nsamp, float *dest );
extern int wfinINTEL3BYTE ( char *buffer, int nbytes, int first, int nsamp, float *dest );
extern int wfout3BYTE ( Wfdisc *wfdisc, Trsample *data, int nsamp, char **outbuf, int *nbytes, int *bufsz );
extern int wfoutINTEL3BYTE ( Wfdisc *wfdisc, Trsample *data, int nsamp, char **outbuf, int *nbytes, int *bufsz );

extern int wfhdrAH ( Wfdisc *wfdisc, void *data, int nsamp, char **header, int *nbytes, int *headersz );
extern int wfhdrCM6 ( Wfdisc *wfdisc, void *data, int nsamp, char **header, int *nbytes, int *headersz );
extern int wfhdrSAC ( Wfdisc * wfdisc, void *data, int nsamp, char **header, int *nbytes, int *headersz, int intel );

extern int trwf2disc ( Wfdisc *wfdisc, char *data, int nbytes, int nsamp );
extern int trwfhdr ( Wfdisc *wfdisc, void *data, int nsamp );
extern int trwftail ( Wfdisc *wfdisc, void *data, int nsamp );

extern int wfinAAH ( char *buffer, int nbytes, int first, int nsamp, float *dest );
extern int wfoutAAH ( Wfdisc *wfdisc, Trsample *data, int nsamp, char **outbuf, int *nbytes, int *bufsz );
extern int wfhdrAAH ( Wfdisc *wfdisc, void *data, int nsamp, char **header, int *nbytes, int *headersz );


extern int trwfdata ( Wfdisc *wfdisc, Trsample *data, int nsamp );
extern int psteim ( FILE *file, char *record, int record_size );

extern int refinit_db ( Dbptr db );
extern int finit_db ( Dbptr db ) ; 
extern int map_seed_netsta ( char *snet, char *fsta, char *sta );
extern int seed_net ( char *sta, char *snet, char *fsta );
extern int map_autodrm_netsta ( char *anet, char *fsta, char *sta );
extern int autodrm_net ( char *sta, char *anet, char *fsta );
extern int map_seed_chanloc ( char *sta, char *fchan, char *loc, char *chan );
extern int seed_loc ( char *sta, char *chan, char *fchan, char *loc );
extern int map_autodrm_chanaux ( char *sta, char *fchan, char *aux, char *chan );
extern int autodrm_aux ( char *sta, char *chan, char *fchan, char *aux );

extern int wtseed (Msd *msd, int n0, int n1) ;
extern int itrwfdata ( Wfdisc *wfdisc, int *data, int nsamp );

extern int msd2wf ( Msd *msd, Dbptr db, int (add_wfdisc) (Dbptr, char *, char *, Wfdisc *) );

#ifdef	__cplusplus
}
#endif

#endif

/* $Id: tr.h,v 1.65 2006/06/20 22:28:11 danq Exp $ */
